This file is part of the Octave test functions for RobCoGen.

Copyright © 2016, 2017, 2018 Marco Frigerio (`marco DOT frigerio AT kuleuven DOT be`)

See the LICENSE file for more information.

---

# Introduction
These Octave/Matlab functions test the code generated by RobCoGen, both C++ and
Octave.

Any test-function compares the numerical output of the RobCoGen implementation
(either C++ or Octave) with the output of the corresponding function in Roy
Featherstone's implementation ([spatial_v2](http://royfeatherstone.org/spatial/v2/),
which is itself written in Octave/Matlab).
The inputs for the function to be tested are randomly generated.

Typically, any test-function would display some numerical results on the screen,
including the difference between the two implementations, which would be zero if
everything is correct.

Roy's code (spatial v2), the RobCoGen-generated Octave code, and obviously the
test-functions we are talking about, must all be in Octave's path. Thus you
might want to run something like the following before attempting to execute the
tests:

```
#!matlab

addpath( genpath( '<your path of spatial v2>' ) );
addpath( '<your path of RobCoGen-generated Octave code>' );
addpath( '<root of the test functions>/src' );
addpath( '<root of the test functions>/src/cpp' );

```

# C++ code
In general, a test-function for the generated C++ code has this form:

`[ME ROY] = testcpp_xxx(MODEL, EXE)`

where `xxx` would change depending on the function (e.g. `id` for inverse
dynamics, `fd_fb` for forward dynamics for floating base robots, etc).

- `MODEL` is the robot model (a data structure) in Roy's format.

- `EXE` is a string with the full path of a C++ executable which shall use the
   RobCoGen dynamics (forward dynamics, inverse dynamics, etc., whichever you
   want to test). The user has to build this executable separately, making sure
   that it satisfies the following assumptions:

    - Command line arguments for all the required inputs, which depend on the
    specific function. The arguments are typically the coordinates
    of the joint status vector, joint velocity vector, etc., in the "usual"
    order.  
    For example, to test the inverse dynamics of a 5-dof fixed-base robot, the
    binary shall take 15 arguments, for the 5 coordinates of the joint position,
    velocity and acceleration vectors.  
    See the documentation or the code of each test-function for specific info.

    - A last command line argument which is the full path of a text file which
    contains the external forces acting on each link (optional). One line for
    each force (i.e. each link), six values for each line (the coordinates of
    the spatial force acting on the link, in link coordinates).

    - The output of the C++ executable must be printouts to stdout. Octave/Matlab
    will get this output and convert it into a local variable to compare it with
    Roy's output. Also the actual output depends on the specific dynamics
    function being tested, obviously. For example, an executable running the
    fixed base forward dynamics of a robot will have to print the joint
    acceleration vector.  
    In general, the results are expected to be printed one after the
    other, as column vectors, one coordinate per line.

- `ME` and `ROY` are data structures with the numerical outputs and comparison
  results.

## Example
In Octave's prompt:
```
#!matlab

roy_model = <robot>();           % RobCoGen-generated Octave code to init Roy's model
exe = '<path to the executable>' % A string with the full path of the binary C++ program

[me roy] = testcpp_jsim_fd(roy_model, exe); % Test the calculation of the JSIM for floating base robots

```

# Octave code
The general signature of a test-function for the generated Octave code is:

`[ME ROY] = test_xxx(ROY_MODEL, ROBCOGEN_MODEL)`

- `ROY_MODEL` is the robot model (a data structure) in Roy's format. Same as the
   C++ test.

- `ROBCOGEN_MODEL` is a data structure with some fields initialized with the
   robot-specific functions generated by RobCoGen. These fields are usually `ip`
   and `xm` or `xf`, which are respectively the inertia properties and the
   coordinate transforms for motion or force vectors. If code was generated
   with no constant folding, than the structure must have another field, `kk`,
   initialized with the generated function `modelConstants()`. Furthermore, if
   the robot model is parametric, a field called `pp` with the desired values
   of the parameters must also be a field of `ROBCOGEN_MODEL`.

- `ME` and `ROY` are data structures with the numerical outputs and comparison
  results.

## Example
In Octave's prompt:

```
#!matlab

roy_model = <robot>();           % RobCoGen-generated Octave code to init Roy's model
rcg.ip = inertiaProperties();    % RobCoGen-generated Octave code with inertia properties
rcg.xm = initMotionTransforms(); % ...

[me roy] = test_id(roy_model, rcg);
```

The same example with code without constant folding:

```
#!matlab
rcg.kk = modelConstants();
rcg.ip = inertiaProperties( rcg.kk );
rcg.xm = initMotionTransforms( rcg.kk );
roy_model = <robot>( rcg.kk );

[me roy] = test_id(roy_model, rcg);
```



# Caveats
**! WARNING !**

RobCoGen code requires the external forces to be supplied in body coordinates.
Roy's code unfortunately expects forces in world coordinates (and it converts
them to body coordinates internally!).

I haven't written a conversion as it is particularly boring.

Therefore the tests as they are now will never succeed, unless you modify
slightly Roy's implementation of the algorithms to expect external forces
already in body coordinates (or you set all forces to be zero). The change is
pretty easy, and it could look like this (example from the floating base
forward dynamics):

```
#!matlab
% ...

% Comment the next line out; that's the original
% pA(6:end) = apply_external_forces(prnt, Xup(6:end), pA(6:end), f_ext(6:end));

for i = 6:model.NB
  pA{i} = pA{i} - f_ext{i}; % everything here is in body coordinates
end

% ...
```


